from copy import deepcopy
import HTMLParser
import re

try:
    from BeautifulSoup import BeautifulSoup as soup
    BS_AVAILABLE = True
except ImportError:
    print "WARNING - BS4 NOT AVAILABLE"
    BS_AVAILABLE = False

class HTMLScraperInterface(object):
    def get_iframes(self, html_in):
        raise NotImplementedError('Subclasses of HTMLScraperInterface should override this method')

    def get_tools(self, html_in):
        raise NotImplementedError('Subclasses of HTMLScraperInterface should override this method')

    def get_assignments(self, html_in):
        raise NotImplementedError('Subclasses of HTMLScraperInterface should override this method')

    def get_grades(self, html_in):
        raise NotImplementedError('Subclasses of HTMLScraperInterface should override this method')

class LXMLParser(HTMLScraperInterface):
    
    def get_iframes(self, html_in):
        doc = soup(html_in)
        frame_attrs = dict(doc.iframe.attrs)
        return [{'name' : frame_attrs['name'],
                 'title': frame_attrs['title'],
                 'src'  : frame_attrs['src'] }]
            

    def get_tools(self, html_in):
        doc = soup(html_in)
        out_dict_list = []
        for tab in doc.findAll('a'):
            class_type = tab.get('class')
            if class_type and re.match('icon-sakai-*', class_type):
                    out_dict_list.append({'name': tab.get('class')[11:].strip(),
                                     'href': tab.get('href'),
                                     'desc': tab.get('title')})
        return out_dict_list
        
    def get_assignments(self, html_in):
        doc = soup(html_in)
        out_list = []
        table = doc.table
        for i, table_row in enumerate(table('tr')):
            if i == 0:
                # skip the table header row
                pass
            else:
                temp_obj = { 'href' : table_row.a['href'] }
                for table_col in table_row('td'):
                    header = table_col.get('headers')
                    if header is not None:
                        temp_obj[header] = table_col.text.strip()
                out_list.append(temp_obj)
        return out_list

    def get_grades(self, html_in):
        doc = soup(html_in)
        out_dict = {}
        tables = doc.findAll('table')
        # tables[0] is the header that we don't care about
        # tables[1] is the course grade
        course_grade_table = tables[1]
        if tables[1].span:
            out_dict['course_grade'] = { 'letter_grade': tables[1].span.text,
                                         'number_grade': tables[1]('span')[1].text }
        else:
            out_dict['course_grade'] = { 'error' : 'Not yet available'}
        out_dict['grades'] = {}
        # tables[2] is a bunch of javascript and all of the grade data.
        # first we have to strip away the javascript.
        row_data = [x for x in tables[2]('td')]
        _NEXT_FIELD = 'name'
        _CURRENT_CATEGORY = ''
        _temp = {}
        for row in row_data:
            if row.img:

                # this is the first row of the table
                continue
            if row.span:
                # this is a category
                _CURRENT_CATEGORY = row.span.text.strip()
                if not _CURRENT_CATEGORY in out_dict['grades']:
                    out_dict['grades'][_CURRENT_CATEGORY] = []
            elif row.get('class') == 'left' and _NEXT_FIELD == 'name':
                # this is a grade name
                _temp['name'] = row.text
                _NEXT_FIELD = 'date'
            elif _NEXT_FIELD == 'date':
                _temp['date'] = row.text
                _NEXT_FIELD = 'grade'
            elif _NEXT_FIELD == 'grade':

                _temp['grade'] = row.text
                _NEXT_FIELD = 'comments'
            elif _NEXT_FIELD == 'comments':
                _temp['comments'] = row.text
                _NEXT_FIELD = 'attachment'
            elif _NEXT_FIELD == 'attachment':
                # ignore this for now
                _NEXT_FIELD = 'name'
                if _CURRENT_CATEGORY == '':
                    if not 'unnamed' in out_dict['grades']:
                        out_dict['grades']['unnamed'] = []
                    out_dict['grades']['unnamed'].append(_temp)
                    _temp = {}
                else:
                    out_dict['grades'][_CURRENT_CATEGORY].append(_temp)
                    _temp = {}
        import pprint; pprint.pprint(out_dict)
        return out_dict

        #     name = row_data[index].text
        #     date = row_data[index + 1].text
        #     grade = row_data[index + 2].text
        #     comments = row_data[index + 3].text
        #     from_assignments = row_data[index + 4].text
        #     out_dict['grades'].append({ 'name' : name,
        #                                 'date' : date,
        #                                 'grade': grade,
        #                                 'comments' : comments,
        #                                 'from_assignments' : from_assignments})
        # return out_dict
        

class DefaultParser(HTMLScraperInterface):
    def get_iframes(self, html_in):
        return _IFrameParser().get_iframes(html_in)

    def get_tools(self, html_in):
        return _SiteToolHTMLParser().get_tools(html_in)

    def get_assignments(self, html_in):
        return _AssignmentHTMLParser().get_assignments(html_in)

        
class _IFrameParser(HTMLParser.HTMLParser):
    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self._iframes = []

    def handle_starttag(self, tag, attrs):
        first_attr = dict(attrs)
        if tag == 'iframe':
            self._iframes.append({ 'name' : first_attr['name'],
                                   'title': first_attr['title'].strip(),
                                   'src'  : first_attr['src']})

    def get_iframes(self, html_input):
        self.feed(html_input)
        return self._iframes
    
    
class _SiteToolHTMLParser(HTMLParser.HTMLParser):
    
    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self._tools = []

    def handle_starttag(self, tag, attrs):
        first_attr = dict(attrs)
        # if this is a link with a class attribute
        if tag == 'a' and 'class' in first_attr:
            # look for tools
            if first_attr['class'].strip() == 'icon-sakai-syllabus':
                self._tools.append({ 'name': 'syllabus',
                                     'href': first_attr['href'],
                                     'desc': first_attr['title']})
            elif first_attr['class'].strip() == 'icon-sakai-resources':
                self._tools.append({ 'name': 'resources',
                                     'href': first_attr['href'],
                                     'desc': first_attr['title']})
            elif first_attr['class'].strip() == 'icon-sakai-assignment-grades':
                self._tools.append({ 'name': 'assignments',
                                     'href': first_attr['href'],
                                     'desc': first_attr['title']})
            elif first_attr['class'].strip() == 'icon-sakai-gradebook-tool':
                self._tools.append({ 'name': 'grades',
                                     'href': first_attr['href'],
                                     'desc': first_attr['title']})

    def get_tools(self, html_text):
        self.feed(html_text)
        return self._tools

    def purge(self):
        self._tools = []

class _AssignmentHTMLParser(HTMLParser.HTMLParser):

    _PARSER_STATE = ['WAITING_FOR_H4',
                     'WAITING_FOR_LINK'
                     'WAITING_FOR_STATUS',
                     'WAITING_FOR_OPEN_DATE',
                     'WAITING_FOR_DUE_DATE']

    _LEXER_STATE = ['STARTING_STATE',
                    'NEXT_IS_TITLE',
                    'NEXT_IS_STATUS',
                    'NEXT_IS_OPEN_DATE',
                    'NEXT_IS_DUE_DATE']

    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self._assignments = []
        self._state = 'WAITING_FOR_H4'
        self._lstate = 'STARTING_STATE'
        self._constructed_obj = {}

    def _assert_state(self, desired_state, desired_lstate):
        return self._state == desired_state and self._lstate == desired_lstate 

    def handle_starttag(self, tag, attr):
        first_attr = dict(attr)
        if tag == 'h4':
            # this is an assignment name
            if self._assert_state('WAITING_FOR_H4', 'STARTING_STATE'):
                self._state = 'WAITING_FOR_LINK'
        elif tag == 'a':
            # this is a link
            if self._assert_state('WAITING_FOR_LINK', 'STARTING_STATE'):
                self._constructed_obj['href'] = first_attr['href']
                self._state = 'WAITING_FOR_TITLE'
                self._lstate = 'NEXT_IS_TITLE'
        elif tag == 'td':
            # this is a table
            if 'headers' in first_attr:
                if first_attr['headers'] == 'status':
                    self._lstate = 'NEXT_IS_STATUS'
                elif first_attr['headers'] == 'openDate':
                    self._lstate = 'NEXT_IS_OPEN_DATE'
                elif first_attr['headers'] == 'dueDate':
                    self._lstate = 'NEXT_IS_DUE_DATE'

    def handle_data(self, data):
        stripped_data = data.strip('\t\n')
        if len(stripped_data) == 0:
            return
        if self._assert_state('WAITING_FOR_TITLE', 'NEXT_IS_TITLE'):
            self._constructed_obj['title'] = stripped_data
            self._state = 'WAITING_FOR_STATUS'
            self._lstate = 'STARTING_STATE'
        elif self._assert_state('WAITING_FOR_STATUS', 'NEXT_IS_STATUS'):
            self._constructed_obj['status'] = stripped_data
            self._state = 'WAITING_FOR_OPEN_DATE'
            self._lstate = 'STARTING_STATE'
        elif self._assert_state('WAITING_FOR_OPEN_DATE', 'NEXT_IS_OPEN_DATE'):
            self._constructed_obj['openDate'] = stripped_data
            self._state = 'WAITING_FOR_DUE_DATE'
            self._lstate = 'STARTING_STATE'
        elif self._assert_state('WAITING_FOR_DUE_DATE', 'NEXT_IS_DUE_DATE'):
            self._constructed_obj['dueDate'] = stripped_data
            self._assignments.append(deepcopy(self._constructed_obj))
            self._constructed_obj = {}
            self._state = self._PARSER_STATE[0]
            self._lstate = self._LEXER_STATE[0]

    def get_assignments(self, html_input):
        self.feed(html_input)
        return self._assignments

    def purge(self):
        self._assignments = []
        self._constructed_obj = {}
        self._state = self._PARSER_STATE[0]
        
REGISTERED_METHODS = { 'default' : DefaultParser,
                       'bs4'     : LXMLParser }


import requests
import parsers

BASE_URL_GATECH = 'https://login.gatech.edu/cas/'
SERVICE = 'https://t-square.gatech.edu/sakai-login-tool/container'
BASE_URL_TSQUARE = 'https://t-square.gatech.edu/direct/'

class TSquareException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class TSquareAuthException(TSquareException):
    pass

class NotAuthenticatedException(TSquareException):
    pass


class SessionExpiredException(TSquareException):
    pass


class AssignmentParseException(TSquareException):
    pass


class TSquareAPI(object):
    def requires_authentication(func):
        """
        Function decorator that throws an exception if the user
        is not authenticated, and executes the function normally
        if the user is authenticated.
        """
        def _auth(self, *args, **kwargs):
            if not self._authenticated:
                raise NotAuthenticatedException('Function {} requires'
                                                .format(func.__name__)
                                                + ' authentication')
            else:
                return func(self, *args, **kwargs)
        return _auth

    def __init__(self, username, password,
                 scraper='bs4'):
        """
        Initialize a TSquareAPI object.
        Logs in to TSquare with username and password.
        @param username - The username to log in with
        @param password - The password to log in with. Not stored.

        @returns A TSquareUser object that represents the user that
                 was logged in.
        @throws TSquareAuthException - If something goes wrong during the
        authentication process (i.e. credentials are bad)
        """
        self._authenticated = True
        self.username = username
        self._tg_ticket, self._service_ticket = _get_ticket(username, password)
        self._session = _tsquare_login(self._service_ticket)
        try:
            self._html_iface = parsers.REGISTERED_METHODS[scraper]()
        except KeyError:
            self._html_iface = parsers.REGISTERED_METHODS['default']()
        

    @requires_authentication
    def logout(self):
        self._session.delete(BASE_URL_GATECH + 'rest/tickets/{}'.format(self._tg_ticket))
        self._authenticated = False
        
    @requires_authentication
    def get_user_info(self):
        """
        Returns a TSquareUser object representing the currently logged in user.
        Throws a NotAuthenticatedException if the user is not authenticated.
        """
        response = self._session.get(BASE_URL_TSQUARE + '/user/current.json')
        response.raise_for_status() # raises an exception if not 200: OK
        user_data = response.json()
        del user_data['password'] # tsquare doesn't store passwords
        return TSquareUser(**user_data)

    @requires_authentication
    def get_sites(self, filter_func=lambda x: True):
        """
        Returns a list of TSquareSite objects that represent the sites available
        to a user.
        @param filter_func - A function taking in a Site object as a parameter
                             that returns a True or False, depending on whether
                             or not that site should be returned by this
                             function. Filter_func should be used to create
                             filters on the list of sites (i.e. user's
                             preferences on what sites to display by default).
                             If not specified, no filter is applied.
        @returns - A list of TSquareSite objects encapsulating t-square's JSON
                   response.
        """
        response = self._session.get(BASE_URL_TSQUARE + 'site.json')
        response.raise_for_status() # raise an exception if not 200: OK
        site_list = response.json()['site_collection']
        if not site_list:
            # this means that this t-square session expired. It's up
            # to the user to re-authenticate.
            self._authenticated = False
            raise SessionExpiredException('The session has expired')
        result_list = []
        for site in site_list:
            t_site = TSquareSite(**site)
            if not hasattr(t_site, "props"):
                t_site.props = {}
            if not 'banner-crn' in t_site.props:
                t_site.props['banner-crn'] = None
            if not 'term' in t_site.props:
                t_site.props['term'] = None
            if not 'term_eid' in t_site.props:
                t_site.props['term_eid'] = None
            if filter_func(t_site):
                result_list.append(t_site)
        return result_list
            
    @requires_authentication
    def get_announcements(self, site=None, num=10, age=20):
        """
        Gets announcements from a site if site is not None, or from every
        site otherwise. Returns a list of TSquareAnnouncement objects.
        @param site_obj (TSquareSite) If non-None, gets only the announcements
                                      from that site. If none, get anouncements
                                      from all sites.
        @param num - The number of announcements to fetch. Default is 10.
        @param age - 'How far back' to go to retreive announcements. Default
                     is 20, which means that only announcements that are
                     less than 20 days old will be returned, even if there
                     less than 'num' of them.
        @returns - A list of TSquareAnnouncement objects. The length will be
                   at most num, and it may be less than num depending on
                   the number of announcements whose age is less than age.
        """
        url = BASE_URL_TSQUARE + 'announcement/'
        if site:
            url += 'site/{}.json?n={}&d={}'.format(site.id, num, age)
        else:
            url += 'user.json?n={}&d={}'.format(num, age)
        request = self._session.get(url)
        request.raise_for_status()
        announcement_list = request.json()['announcement_collection']
        return map(lambda x: TSquareAnnouncement(**x), announcement_list)

    @requires_authentication
    def get_tools(self, site):
        """
        Gets all tools associated with a site.
        @param site (TSquareSite) - The site to search for tools
        @returns A list of dictionaries representing Tsquare tools.
        """
        # hack - gotta bypass the tsquare REST api because it kinda sucks with tools
        url = site.entityURL.replace('direct', 'portal')
        response = self._session.get(url)
        response.raise_for_status()
        # scrape the resulting html
        tools_dict_list = self._html_iface.get_tools(response.text)
        return [TSquareTool(**x) for x in tools_dict_list]

    @requires_authentication
    def get_assignments(self, site):
        """
        Gets a list of assignments associated with a site (class). Returns
        a list of TSquareAssignment objects.
        @param site (TSquareSite) - The site to use with the assignment query

        @returns - A list of TSquareSite objects. May be an empty list if
                   the site has defined no assignments.
        """
        tools = self.get_tools(site)
        assignment_tool_filter = [x for x in tools if x.name == 'assignment-grades']
        if not assignment_tool_filter:
            return []
        assignment_tool_url = assignment_tool_filter[0].href
        response = self._session.get(assignment_tool_url)
        response.raise_for_status()
        iframes = self._html_iface.get_iframes(response.text)
        iframe_url = ''
        for frame in iframes:
            if frame['title'] == 'Assignments ':
                iframe_url = frame['src']
        if iframe_url == '':
            print "WARNING: NO ASSIGNMENT IFRAMES FOUND"
        response = self._session.get(iframe_url)
        response.raise_for_status()
        assignment_dict_list = self._html_iface.get_assignments(response.text)
        return [TSquareAssignment(**x) for x in assignment_dict_list]

    @requires_authentication
    def get_grades(self, site):
        tools = self.get_tools(site)
        grade_tool_filter = [x for x in tools if x.name == 'gradebook-tool']
        if not grade_tool_filter:
            return []
        response = self._session.get(grade_tool_filter[0])
        response.raise_for_status()
        iframes = self._html_iface.get_iframes(response.text)
        iframe_url = ''
        for frame in iframes:
            if frame['title'] == 'Gradebook ':
                iframe_url = frame['src']
        if iframe_url == '':
            print "WARNING: NO GRADEBOOK IFRAMES FOUND"
        response = self._session.get(iframe_url)
        response.raise_for_status()
        grade_dict_list = self._html_iface.get_grades(response.text)
        return grade_dict_list
        
        
class TSquareUser:
    def __init__(self, **kwargs):
        """
        Encapsulates the raw JSON dictionary that represents a user in TSquare.
        Converts a dictionary to attributes of an object for ease of use.
        This constructor should never be called directly; instead, it is
        called by get_user_info.
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])

        def list_attrs(self):
            return self.__dict__.keys()

class TSquareSite:
    def __init__(self, **kwargs):
        """
        Encapsulates the raw JSON dictionary that represents a site in TSquare.
        Converts a dictionary to attributes of an object for ease of use.
        This constructor should never be called directly; instead, it is called
        by get_sites.
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])

class TSquareAnnouncement:
    def __init__(self, **kwargs):
        """
        Encapsulates the raw JSON dictionary that represents an announcement
        in TSquare.
        Converts a dictionary to attributes of an object for ease of use.
        This constructor should never be called directly; instead, it is called
        by get_announcements.
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])

class TSquareTool:
    def __init__(self, **kwargs):
        """
        Encapsulates the raw JSON dictionary that represents a tool in TSquare.
        A tool is any third party application that TSquare uses to provide a
        service. In this case, assignments, grades, and resources are the most
        common tools in use.
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])

class TSquareAssignment:
    def __init__(self, **kwargs):
        """
        Encapsulates the dictionary that this module builds by scraping the
        Assignments page. An assignment is anything that can be turned
        in according to TSquare.
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])
        
def _get_ticket(username, password):
    # step 1 - get a CAS ticket
    data = { 'username' : username, 'password' : password }
    response = requests.post(BASE_URL_GATECH + 'rest/tickets', data=data)
    if response.status_code == 400:
        raise TSquareAuthException('Username or password incorrect')
    elif not response.status_code == 201:
        raise TSquareAuthException('Received unexpected HTTP code: {}'
                                   .format(response.status_code))
    # black magic to strip the ticket out of the raw html response
    form_split = response.text.split('<form action="')[1].split(' ')[0]
    ticket = form_split.split('tickets/')[1][:-1]
    # step 2 - get a TSquare service ticket
    data = { 'service' : SERVICE }
    response = requests.post(BASE_URL_GATECH + 'rest/tickets/{}'.format(ticket),
                             data=data)
    if response.status_code == 400:
        raise TSquareAuthException('Parameters missing from ST call')
    elif not response.status_code == 200:
        raise TSquareAuthException('Received unexpected HTTP code: {}'
                                   .format(response.status_code))
    service_ticket = response.text
    return ticket, service_ticket


def _tsquare_login(service_ticket):
    session = requests.Session()
    # step 3 - redeem the ticket with TSquare and receive authenticated session
    session.get(SERVICE + '?ticket={}'.format(service_ticket))
    return session





try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name="tsquare",
      version='0.2.1',
      description='Python TSquare API bindings',
      author='Sean Gillespie',
      author_email='sean.william.g@gmail.com',
      url='https://github.com/swgillespie/tsquare',
      py_modules=['tsquare',
                  'tsquare.core',
                  'tsquare.parsers'],
      long_description="Get and manipulate the state of TSquare with python!",
      install_requires=['requests>=1.2.3',
                        'BeautifulSoup>=3.2.1',
                        'lxml>=3.1.0'],
      license='MIT',
      )